import os
from pathlib import Path
from typing import Text

import pytest
from rasa.constants import DEFAULT_DATA_PATH

from rasa.nlu.training_data.formats import MarkdownReader, RasaReader
from rasa.nlu.training_data.loading import MARKDOWN, RASA

from rasax.community import config
from rasax.community.services.data_service import DataService
from rasax.community.services.domain_service import DomainService
from rasax.community.services.intent_service import INTENT_EXAMPLES_KEY
from rasax.community.services.story_service import StoryService
from tests.unit.conftest import NLU_DATA_PATH, _user_event
from tests.unit.conftest import TEST_USER, user_dict

TEST_PROJECT = config.project_name

TEST_EXAMPLE = {
    "text": "a great example",
    "intent": "testing",
    "entities": [{"start": 0, "end": 1, "value": "indefinite", "entity": "article"}],
    "team": "team",
    "project_id": "my_project",
}

TEST_RESULT = {
    "id": 1,
    "text": TEST_EXAMPLE["text"],
    "entities": TEST_EXAMPLE["entities"],
    "intent": TEST_EXAMPLE["intent"],
    "hash": "51601e6f1383c5bc88ff25bfaa080a20",
    "annotation": {"user": TEST_USER},
}

EXAMPLE_DATA_MD = """
## intent:check_balance
- what is my balance
- how much do I have on my [savings](source_account)
- how much do I have on my [savings account](source_account:savings)
- Could I pay in [yen](currency)?

## intent:greet
- hey
- hello

## synonym:savings
- pink pig

## regex:zipcode
- [0-9]{5}

## lookup:currencies
- Yen
- USD
- Euro
"""

TEST_LOOKUP_TABLE = """
Berlin
Hamburg
Frankfurt
""".strip()


def test_save_example(data_service: DataService):
    save_result = data_service.save_example(
        user_dict()["username"], TEST_PROJECT, TEST_EXAMPLE
    )

    actual, number_examples = data_service.get_training_data(TEST_PROJECT)

    assert number_examples == 1
    assert save_result == actual[0]
    assert actual[0]["annotation"].pop("time") > 0
    assert actual[0] == TEST_RESULT


def test_save_example_if_it_already_exists(data_service: DataService):
    updated_example = TEST_EXAMPLE.copy()
    updated_intent = "another intent"
    updated_example["intent"] = updated_intent

    data_service.save_example(user_dict()["username"], TEST_PROJECT, updated_example)

    expected = TEST_RESULT.copy()
    expected["intent"] = updated_intent

    actual, number_examples = data_service.get_training_data(TEST_PROJECT)

    assert number_examples == 1
    assert actual[0]["annotation"].pop("time") > 0
    assert actual[0] == expected


def test_get_by_hash(data_service: DataService):
    data_service.save_example(user_dict()["username"], TEST_PROJECT, TEST_EXAMPLE)
    actual = data_service.get_example_by_hash(TEST_PROJECT, TEST_RESULT["hash"])

    assert actual["annotation"].pop("time") > 0
    assert actual == TEST_RESULT


def test_get_by_intent_query(data_service: DataService):
    data_service.save_example(user_dict()["username"], TEST_PROJECT, TEST_EXAMPLE)
    intent_query = "some intent,other_intent,{}".format(TEST_EXAMPLE["intent"])

    actual, _ = data_service.get_training_data(TEST_PROJECT, intent_query=intent_query)

    assert actual[0]["annotation"].pop("time") > 0
    assert actual[0] == TEST_RESULT


def test_get_with_text_query(data_service: DataService):
    text_query = "great"

    actual, _ = data_service.get_training_data(TEST_PROJECT, text_query=text_query)

    assert actual[0]["annotation"].pop("time") > 0
    assert actual[0] == TEST_RESULT


def test_first_level_field_query(data_service: DataService):
    actual, _ = data_service.get_training_data(
        TEST_PROJECT, fields_query=[("text", True)]
    )
    assert actual[0] == {"text": TEST_EXAMPLE["text"]}


def test_relationship_field_query(data_service: DataService):
    actual, _ = data_service.get_training_data(
        TEST_PROJECT, fields_query=[("entities", True)]
    )
    assert actual[0] == {"entities": TEST_EXAMPLE["entities"]}


def test_relationship_subfield_query(data_service: DataService):
    actual, _ = data_service.get_training_data(
        TEST_PROJECT, fields_query=[("entities.entity", True)]
    )
    assert actual[0] == {"entities": TEST_EXAMPLE["entities"]}


def test_get_intents(data_service: DataService):
    actual = data_service.get_intents(TEST_PROJECT)

    assert actual == [
        {"intent": TEST_RESULT["intent"], INTENT_EXAMPLES_KEY: [TEST_RESULT["hash"]]}
    ]


def test_get_intents_without_example_hashes(data_service: DataService):
    actual = data_service.get_intents(TEST_PROJECT, include_example_hashes=False)

    assert actual == [{"intent": TEST_RESULT["intent"]}]


def test_update_intent(data_service: DataService):
    new_intent = "new intent"
    data_service.update_intent(new_intent, [TEST_RESULT["hash"]], TEST_PROJECT)

    assert data_service.get_intents(TEST_PROJECT) == [
        {"intent": new_intent, INTENT_EXAMPLES_KEY: [TEST_RESULT["hash"]]}
    ]


def test_replace_example(data_service: DataService):
    updated_example = TEST_EXAMPLE.copy()
    updated_example["text"] = "another text"

    data_service.replace_example(
        user_dict(), TEST_PROJECT, updated_example, TEST_RESULT["id"]
    )

    expected = TEST_RESULT.copy()
    expected["text"] = "another text"
    expected["hash"] = "2275675a128d775f75ea2c4f3ba6be46"

    actual, number_examples = data_service.get_training_data(TEST_PROJECT)

    assert number_examples == 1
    assert actual[0]["annotation"].pop("time") > 0
    assert actual[0] == expected


def test_get_entities(data_service: DataService):
    assert data_service.get_entities(TEST_PROJECT) == [
        {"entity": TEST_RESULT["entities"][0]["entity"]}
    ]


def test_get_warnings(data_service: DataService):
    expected = [
        {"type": "data", "min": 4, "count": 1},
        {"type": "intent", "min": 2, "count": 1, "meta": None},
        {"type": "dataPerIntent", "min": 2, "count": 1, "meta": "testing"},
    ]

    assert data_service.get_training_data_warnings(TEST_PROJECT) == expected


def test_delete_example(data_service: DataService):
    actual = data_service.delete_example(TEST_RESULT["id"])

    assert actual
    assert data_service.get_training_data(TEST_PROJECT) == ([], 0)


def test_delete_example_by_hash(data_service: DataService):
    data_service.save_example(user_dict()["username"], TEST_PROJECT, TEST_EXAMPLE)

    assert data_service.delete_example_by_hash(TEST_PROJECT, TEST_RESULT["hash"])
    assert data_service.get_training_data(TEST_PROJECT) == ([], 0)


def test_replace_additional_training_features(
    data_service: DataService, new_project: Text
):
    training_data = MarkdownReader().reads(EXAMPLE_DATA_MD)
    data_service.replace_additional_training_features(new_project, training_data)

    assert data_service.get_regex_features(new_project).result == [
        {"id": 1, "name": "zipcode", "pattern": "[0-9]{5}"}
    ]
    assert data_service.get_lookup_tables(new_project) == [
        {
            "id": 1,
            "name": "currencies",
            "elements": os.path.join(DEFAULT_DATA_PATH, "currencies"),
            "number_of_elements": 3,
        }
    ]

    synonyms = data_service.get_entity_synonyms(new_project)
    assert synonyms[0]["value"] == "savings"
    assert set(synonyms[0]["synonyms"]) == {"pink pig", "savings account"}


def test_store_lookup_table(data_service: DataService, new_project: Text):
    filename = "test"
    created = data_service.save_lookup_table(filename, TEST_LOOKUP_TABLE, new_project)

    actual = data_service.get_lookup_tables(new_project)
    assert {
        "id": created["id"],
        "name": filename,
        "elements": os.path.join(DEFAULT_DATA_PATH, filename),
        "number_of_elements": 3,
    } in actual

    assert data_service.get_lookup_table(created["id"]) == TEST_LOOKUP_TABLE


def test_get_lookup_tables_with_elements(data_service: DataService, new_project: Text):
    filename1 = "test.txt"
    data_service.save_lookup_table(filename1, TEST_LOOKUP_TABLE, new_project)

    lookup_table2 = """
        Chicago
        SF
        """.strip()

    filename2 = "test2.txt"
    data_service.save_lookup_table(filename2, lookup_table2, new_project)

    actual = data_service.get_lookup_tables_with_elements(new_project)

    assert actual == [
        {"name": filename1, "elements": ["Berlin", "Hamburg", "Frankfurt"]},
        {"name": filename2, "elements": ["Chicago", "SF"]},
    ]


def test_delete_lookup_table(data_service: DataService, new_project: Text):
    lookup_table = ""
    filename = "test"
    created = data_service.save_lookup_table(filename, lookup_table, new_project)

    data_service.delete_lookup_table(created["id"])

    still_existing = data_service.get_lookup_tables(new_project)
    assert created not in still_existing


def test_delete_lookup_table_with_invalid_id(data_service: DataService):
    with pytest.raises(ValueError):
        data_service.delete_lookup_table(3456789)


def test_add_regex_feature(data_service: DataService, new_project: Text):
    training_example = {"name": "test_add_regex_feature", "pattern": ".*"}
    created = data_service.create_regex_feature(training_example, new_project)

    assert created["id"]
    actual = data_service.get_regex_feature_by_id(created["id"])
    actual.pop("id")

    assert actual == training_example


def test_add_regex_feature_with_same_pattern(
    data_service: DataService, new_project: Text
):
    training_example = {
        "name": "test_add_regex_feature_with_same_pattern",
        "pattern": ".*",
    }
    _ = data_service.create_regex_feature(training_example, new_project)

    with pytest.raises(ValueError):
        data_service.create_regex_feature(training_example, new_project)


def test_update_regex(data_service: DataService, new_project: Text):
    training_example = {
        "name": "test_add_regex_feature_with_same_pattern",
        "pattern": ".*",
    }
    created = data_service.create_regex_feature(training_example, new_project)

    updated = {"id": created["id"], "name": "so updated", "pattern": "sdadas"}

    assert data_service.update_regex_feature(created["id"], updated) == updated
    assert data_service.get_regex_feature_by_id(created["id"]) == updated


def test_update_regex_with_invalid_id(data_service: DataService):
    with pytest.raises(ValueError):
        data_service.update_regex_feature(9929, {})


def test_delete_regex(data_service: DataService, new_project: Text):
    created = data_service.create_regex_feature(
        {"name": "test_delete_endpoint", "pattern": ".*"}, new_project
    )

    data_service.delete_regex_feature(created["id"])

    with pytest.raises(ValueError):
        data_service.get_regex_feature_by_id(created["id"])


def test_delete_regex_with_invalid_id(data_service: DataService):
    with pytest.raises(ValueError):
        data_service.delete_regex_feature(9929)


def test_format_entity_synonyms(data_service: DataService):
    # insert example data
    training_data = MarkdownReader().reads(EXAMPLE_DATA_MD)
    data_service.replace_additional_training_features(TEST_PROJECT, training_data)

    # get entity synonyms and ensure RasaReader reads them correctly
    formatted_synonyms = data_service.get_entity_synonyms(TEST_PROJECT)
    assert formatted_synonyms
    training_data_json = {"rasa_nlu_data": {"entity_synonyms": formatted_synonyms}}
    training_data = RasaReader().read_from_json(training_data_json)
    assert training_data.entity_synonyms == {
        "savings account": "savings",
        "pink pig": "savings",
    }


def test_domain_items_from_nlu_data(
    data_service: DataService, domain_service: DomainService
):
    from tests.unit.services.test_domain_service import TEST_DOMAIN

    # ensure there is a domain
    domain_service.store_domain(
        TEST_DOMAIN, TEST_PROJECT, username=TEST_USER, store_templates=False
    )

    # add an example
    data_service.save_example(
        TEST_USER, TEST_PROJECT, TEST_EXAMPLE, add_example_items_to_domain=False
    )
    data_service.add_all_intents_and_entities_to_domain(TEST_PROJECT)

    domain_intents = domain_service.get_intents_from_domain(TEST_PROJECT)
    domain_entities = domain_service.get_entities_from_domain(TEST_PROJECT)

    # ensure domain contains example's entity and intent
    assert TEST_EXAMPLE["intent"] in domain_intents
    assert TEST_EXAMPLE["entities"][0]["entity"] in domain_entities


async def test_domain_items_from_story(
    story_service: StoryService, domain_service: DomainService
):
    from tests.unit.services.test_domain_service import TEST_DOMAIN

    # ensure there is a domain
    domain_service.store_domain(
        TEST_DOMAIN, TEST_PROJECT, username=TEST_USER, store_templates=False
    )

    # add a story
    intent = "some_intent1"
    action = "some_action1"
    story = """
    ## story1
    * {}
      - {}
    """.format(
        intent, action
    )
    await story_service.save_stories(
        story,
        "team",
        username=TEST_USER,
        dump_stories_in_local_mode=False,
        add_story_items_to_domain=True,
    )

    domain_actions = domain_service.get_actions_from_domain(TEST_PROJECT)
    domain_intents = domain_service.get_intents_from_domain(TEST_PROJECT)

    # ensure domain contain story action and intent
    assert action in domain_actions
    assert intent in domain_intents


def test_inject_data_from_file(data_service: DataService, new_project: Text):
    assert not data_service.get_training_data(new_project)[0]
    expected = MarkdownReader().read(NLU_DATA_PATH)

    # insert data from file
    data_service.save_bulk_data_from_files([NLU_DATA_PATH], new_project, TEST_USER)
    assert data_service.get_training_data(new_project)[0]

    actual = data_service.get_nlu_training_data_object(project_id=new_project)

    # use markdown to test equality since json can contain ids from the database
    assert actual.training_examples == expected.training_examples


def test_inject_data_with_invalid_file_ending(
    data_service: DataService, tmp_path: Path
):
    file = tmp_path / "my_text.txt"
    file.write_text("test test test test")
    inserted = data_service.save_bulk_data_from_files([str(file)])
    assert inserted.is_empty()


def test_entity_extractor_persistence(data_service: DataService):
    extractor = "some_extractor"
    test_example_with_extractor = TEST_EXAMPLE.copy()
    test_example_with_extractor["entities"][0]["extractor"] = extractor
    # modify text so it's saved as a new example
    new_text = "new_text"
    test_example_with_extractor["text"] = new_text

    # save example
    save_result = data_service.save_example(
        user_dict()["username"], TEST_PROJECT, test_example_with_extractor
    )

    # make sure extractor was saved
    actual, number_examples = data_service.get_training_data(
        TEST_PROJECT, text_query=new_text
    )
    assert number_examples == 1
    assert save_result == actual[0]
    assert actual[0]["entities"][0]["extractor"] == extractor


@pytest.mark.parametrize(
    "training_data, format",
    [
        (MarkdownReader().reads(EXAMPLE_DATA_MD).as_markdown(), MARKDOWN),
        (MarkdownReader().reads(EXAMPLE_DATA_MD).as_json(), RASA),
    ],
)
def test_replace_data(
    data_service: DataService, new_project: Text, training_data: Text, format: Text
):
    data_service.replace_data(new_project, training_data, format)

    assert data_service.get_training_data(new_project)


def test_save_bulk_data_with_unknown_format(data_service: DataService):
    assert not data_service.save_bulk_data("project", "some data", "unknown format")


def test_get_all_filenames(data_service: DataService, new_project: Text):
    filenames = {"one", "two"}
    for file in filenames:
        example = TEST_EXAMPLE.copy()
        example["text"] = file
        _ = data_service.save_example(
            user_dict()["username"], new_project, example, filename=file
        )

    assert set(data_service.get_all_filenames(new_project)) == filenames

    assert all(
        [
            len(data_service.get_training_data_for_filename(new_project, f)) == 1
            for f in filenames
        ]
    )


def test_save_user_event_as_example(data_service: DataService, new_project: Text):
    test_text = "Ready to launch"
    test_intent = "test_save_user_event_as_example"
    event = _user_event(
        "sender", intent=test_intent, text=test_text, entities=TEST_EXAMPLE["entities"]
    )
    data_service.save_user_event_as_example(user_dict(), new_project, event)

    actual, number_of_results = data_service.get_training_data(new_project)

    assert number_of_results == 1
    assert actual[0]["text"] == test_text
    assert actual[0]["intent"] == test_intent
    assert actual[0]["entities"] == TEST_EXAMPLE["entities"]


def test_dump_data_in_local_mode(data_service: DataService, new_project: Text):
    config.LOCAL_MODE = True
    filenames = {"dir/one.md", "dir/two.md"}
    for file in filenames:
        example = TEST_EXAMPLE.copy()
        example["text"] = file
        _ = data_service.save_example(
            user_dict()["username"], new_project, example, filename=file
        )

    data_service._dump_data_in_local_mode(new_project)

    assert all([os.path.exists(f) for f in filenames])
    config.LOCAL_MODE = False
